package fr.lri.hcc.ilda.touchtokens;

/*
 *   Authors: Caroline Appert (caroline.appert@lri.fr)
 *   Copyright (c) Univ. Paris-Sud XI, 2016. All Rights Reserved
 *   Licensed under the GNU LGPL. For full terms see the file COPYING.
*/

import android.content.Context;
import android.support.v4.view.MotionEventCompat;
import android.util.AttributeSet;
import android.util.Log;
import android.view.MotionEvent;
import android.view.View;

import java.util.ArrayList;
import java.util.Timer;
import java.util.TimerTask;

/**
 * Created by appert on 02/09/15.
 */
public abstract class TouchTokenView extends View {

    protected String tokenDown = null;

    protected static TouchPoint[] FINGERS_LOCATION = new TouchPoint[10];
    protected static TouchPoint[] FINGERS_LOCATION_AT_TIMER_START_TIME = new TouchPoint[10];

    protected int dwellDuration = 200; // in ms
    protected double dwellTolerance = 3; // in mm

    protected Timer dwellTimer;
    protected long dwellStartTime = -1;
    protected long lastUpdateTime = -1;

    protected boolean debug = false;

    protected double viewWidthMm = -1;
    protected double viewHeightMm = -1;

    protected ArrayList<TouchPoint> notchPointsAtDown;
    protected TokenTemplate currentTemplate;

    public TouchTokenView(Context context) {
        super(context);
        init();
    }

    public TouchTokenView(Context context, AttributeSet attrs) {
        super(context, attrs);
        init();
    }

    public TouchTokenView(Context context, AttributeSet attrs, int defStyleAttr) {
        super(context, attrs, defStyleAttr);
        init();
    }

    protected void init() {


        viewWidthMm = getWidth() / MainActivity.SCREEN_RESOLUTION;
        viewHeightMm = getHeight() / MainActivity.SCREEN_RESOLUTION;

        System.out.println("getWidth() " + getWidth());
        System.out.println("getHeight() " + getHeight());
        System.out.println("MainActivity.SCREEN_RESOLUTION " + MainActivity.SCREEN_RESOLUTION);

        dwellTimer = new Timer();
        dwellTimer.schedule(new TimerTask() {
            public void run() {
                if (dwellStartTime == -1) {
                    return;
                }
                if (tokenDown != null) {
                    return;
                }
                long currentTime = System.currentTimeMillis();
                if ((currentTime - dwellStartTime) >= dwellDuration && (lastUpdateTime - dwellStartTime) < dwellDuration) {
                    ArrayList<TouchPoint> inputInMms = new ArrayList<TouchPoint>(); // the recognizer stores templates in mm
                    ArrayList<TouchPoint> inputIn01 = new ArrayList<TouchPoint>();
                    for (int i = 0; i < FINGERS_LOCATION.length; i++) {
                        if (FINGERS_LOCATION[i] != null) {
                            inputInMms.add(new TouchPoint(FINGERS_LOCATION[i].x * viewWidthMm, FINGERS_LOCATION[i].y * viewHeightMm));
                            inputIn01.add(new TouchPoint(FINGERS_LOCATION[i].x, FINGERS_LOCATION[i].y));
                        }
                    }
                    if (debug) {
                        System.out.println("input: " + inputInMms);
                    }
                    currentTemplate = MainActivity.RECOGNIZER.recognize(inputInMms);

                    if (debug) {
                        System.out.println("recognized as: " + currentTemplate);
                    }
                    if (currentTemplate != null) {
                        tokenDown = currentTemplate.getTokenID();
                        notchPointsAtDown = new ArrayList<TouchPoint>();
                        for (TouchPoint pt : currentTemplate.getPoints()) {
                            notchPointsAtDown.add(new TouchPoint(pt.x / viewWidthMm, pt.y / viewHeightMm));
                        }
                        ArrayList<TouchPoint> tokenPoints = new ArrayList<TouchPoint>(inputIn01);
                        onTokenDown(tokenDown, inputIn01, tokenPoints, viewWidthMm, viewWidthMm, currentTemplate);
                    }
                }
                lastUpdateTime = System.currentTimeMillis();
            }
        }, 0, 25);

    }

    protected void complementNotchPoints(ArrayList<TouchPoint> notchPointsAtDown, ArrayList<TouchPoint> touchPoints, double deviceWidth, double deviceHeight) {
        Log.v("DEBUG", "notchPointsAtDown.size()=" + notchPointsAtDown.size() + ", touchPoints.size()=" + touchPoints.size());
        int missingIndex = -1;
        TouchPoint notchDown1 = null;
        TouchPoint notch1 = null;
        TouchPoint notchDown2 = null;
        TouchPoint notch2 = null;
        TouchPoint notchDown3 = null;
        for (int i = 0; i < touchPoints.size(); i++) {
            if(touchPoints.get(i) == null) {
                missingIndex = i;
                notchDown3 = notchPointsAtDown.get(i);
                notchDown3 = new TouchPoint(notchDown3.x*deviceWidth, notchDown3.y*deviceHeight);
            } else {
                if(notchDown1 == null) {
                    notchDown1 = notchPointsAtDown.get(i);
                    notchDown1 = new TouchPoint(notchDown1.x*deviceWidth, notchDown1.y*deviceHeight);
                    notch1 = touchPoints.get(i);
                    notch1 = new TouchPoint(notch1.x*deviceWidth, notch1.y*deviceHeight);
                } else {
                    notchDown2 = notchPointsAtDown.get(i);
                    notchDown2 = new TouchPoint(notchDown2.x*deviceWidth, notchDown2.y*deviceHeight);
                    notch2 = touchPoints.get(i);
                    notch2 = new TouchPoint(notch2.x*deviceWidth, notch2.y*deviceHeight);
                }
            }
        }
        if(missingIndex == -1) {
            return;
        } else {
            TouchPoint vNotchesDown13 = new TouchPoint(notchDown3.x - notchDown1.x, notchDown3.y - notchDown1.y);
            TouchPoint vNotchesDown12 = new TouchPoint(notchDown2.x - notchDown1.x, notchDown2.y - notchDown1.y);
            double angleNotchesDown213 = TouchTokenRecognizer.angleBetweenVectors(vNotchesDown12, vNotchesDown13);

            TouchPoint vNotches12 = new TouchPoint(notch2.x - notch1.x, notch2.y - notch1.y);
            double angleNotchesx12 = TouchTokenRecognizer.angleBetweenVectors(new TouchPoint(1, 0), vNotches12);

            double lengthNotchesDown13 =
                    Math.sqrt((notchDown3.x - notchDown1.x) * (notchDown3.x - notchDown1.x) +
                            (notchDown3.y - notchDown1.y) * (notchDown3.y - notchDown1.y));

            TouchPoint notch3 = new TouchPoint(lengthNotchesDown13, 0);
            notch3.setLocation(notch3.x + notch1.getX(), notch3.y + notch1.getY());
            double angle = angleNotchesx12 + angleNotchesDown213;
            TouchTokenRecognizer.rotateBy(notch3, notch1, angle, notch3);
            notch3.setLocation(notch3.x / deviceWidth, notch3.y / deviceHeight);
            touchPoints.set(missingIndex, notch3);
        }
    }




    public abstract void onTokenDown(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate);

    public abstract void onTokenMoved(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate);

    public abstract void onTokenUp(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate);

    public void fingerDown(MotionEvent event, int index, int id) {
        Log.v("TOUCHTOKENS", "index="+index+", id="+id);
        FINGERS_LOCATION[id] = new TouchPoint(
                event.getX(index)/getWidth(),
                event.getY(index)/getHeight());
        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            FINGERS_LOCATION_AT_TIMER_START_TIME[i] = FINGERS_LOCATION[i];
        }
        dwellStartTime = System.currentTimeMillis();
    }

    public void fingerMove(MotionEvent event, int index, int id) {
        FINGERS_LOCATION[id] = new TouchPoint(
                event.getX(index)/getWidth(),
                event.getY(index)/getHeight());

        // sometimes, MT events are not delivered in the right order, avoid null exception:
        if(FINGERS_LOCATION_AT_TIMER_START_TIME[id] == null) {
            FINGERS_LOCATION_AT_TIMER_START_TIME[id] = FINGERS_LOCATION[id];
        }

        TouchPoint pointNow = new TouchPoint(
                FINGERS_LOCATION[id].x * viewWidthMm,
                FINGERS_LOCATION[id].y * viewHeightMm);
        TouchPoint pointAtTimerStartTime = new TouchPoint(
                FINGERS_LOCATION_AT_TIMER_START_TIME[id].x * viewWidthMm,
                FINGERS_LOCATION_AT_TIMER_START_TIME[id].y * viewHeightMm);

        double distance = pointNow.distance(pointAtTimerStartTime);

        boolean dwelling = distance < dwellTolerance;
        if(!dwelling) {
            for (int i = 0; i < FINGERS_LOCATION.length; i++) {
                FINGERS_LOCATION_AT_TIMER_START_TIME[i] = FINGERS_LOCATION[i];
            }
            dwellStartTime = System.currentTimeMillis();
        }

        ArrayList<TouchPoint> inputIn01 = new ArrayList<TouchPoint>();
        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            if(FINGERS_LOCATION[i] != null) {
                inputIn01.add(new TouchPoint(FINGERS_LOCATION[i].x, FINGERS_LOCATION[i].y));
            } else {
                inputIn01.add(null);
            }
        }
        if(tokenDown != null) {
            ArrayList<TouchPoint> notchPoints = new ArrayList<TouchPoint>();
            int pointsCount = 0;
            for (int i = 0; i < 3; i++) {
                TouchPoint touchPoint = inputIn01.get(i);
                notchPoints.add(touchPoint);
                if(touchPoint != null) {
                    pointsCount++;
                }
            }
            if(pointsCount == 1) {
                tokenUp();
            }
            if(pointsCount == 2) {
                complementNotchPoints(notchPointsAtDown, notchPoints, viewWidthMm, viewHeightMm);
            }
            if(pointsCount == 2 || pointsCount == 3) {
                onTokenMoved(tokenDown, inputIn01, notchPoints, viewWidthMm, viewHeightMm, currentTemplate);
            }
        }
    }

    protected void tokenUp() {
        if(tokenDown != null) {
            dwellStartTime = -1;
            ArrayList<TouchPoint> inputIn01 = new ArrayList<TouchPoint>();
            for (int i = 0; i < FINGERS_LOCATION.length; i++) {
                if(FINGERS_LOCATION[i] != null) {
                    inputIn01.add(new TouchPoint(FINGERS_LOCATION[i].x, FINGERS_LOCATION[i].y));
                }
            }
            onTokenUp(tokenDown, inputIn01, inputIn01, viewWidthMm, viewHeightMm, currentTemplate);
            tokenDown = null;
        }
    }

    public void fingerUp(MotionEvent event, int index, int id) {
        int cursorID = id;
        int fingerCount = 0;
        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            if(FINGERS_LOCATION[i] != null) {
                fingerCount++;
            }
        }
        if(fingerCount == 1) {
            tokenUp();
        } else {
            dwellStartTime = System.currentTimeMillis();
        }
        FINGERS_LOCATION[cursorID] = null;
        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            FINGERS_LOCATION_AT_TIMER_START_TIME[i] = FINGERS_LOCATION[i];
        }
    }

    public boolean onTouchEvent(MotionEvent event) {
        int index = MotionEventCompat.getActionIndex(event);
        int action = MotionEventCompat.getActionMasked(event);
        int id = MotionEventCompat.getPointerId(event, index);
        if(debug) {
            Log.v("TOUCHTOKENS", action+" ("+event.getX(index)+", "+event.getY(index)+") "+id+" - "+index);
        }
        switch (action) {
            case MotionEvent.ACTION_DOWN:
                fingerDown(event, index, id);
                break;
            case MotionEvent.ACTION_POINTER_DOWN:
                fingerDown(event, index, id);
                break;
            case MotionEvent.ACTION_MOVE:
                //fingerMove(event, index, id);
                int pointerCount = event.getPointerCount();
                for(int i = 0; i < pointerCount; ++i) {
                    index = i;
                    id = MotionEventCompat.getPointerId(event, index);
                    if(id < 10){
                        fingerMove(event, index, id);
                    }
                }
                break;
            case MotionEvent.ACTION_UP:
                fingerUp(event, index, id);
                break;
            case MotionEvent.ACTION_POINTER_UP:
                fingerUp(event, index, id);
                break;
            default:
                return true;
        }
        if(debug) {
            String points = "";
            for (int i = 0; i < FINGERS_LOCATION.length; i++) {
                if(FINGERS_LOCATION[i] != null) {
                    points += " "+FINGERS_LOCATION[i];
                }
            }
            Log.v("TOUCHTOKENS", "points="+points);
        }
        invalidate();
        return true;
    }

    public double getTokenInitOrientation(TokenTemplate transformedTemplate) {
        return transformedTemplate.getOrientation();
    }

    public double getTokenRelOrientation(TokenTemplate transformedTemplate, ArrayList<TouchPoint> notchPoints) {
        TouchPoint transformedTemplateCenter = transformedTemplate.getCenter();
        TouchPoint transformedTemplate1 = transformedTemplate.getPoints().get(0);
        TouchPoint vTransformedTemplate = new TouchPoint(
                transformedTemplate1.x - transformedTemplateCenter.x,
                transformedTemplate1.y - transformedTemplateCenter.y);
        TouchPoint currentCenter = getTokenCenter(transformedTemplate, notchPoints);
        currentCenter = new TouchPoint(currentCenter.x*viewWidthMm, currentCenter.y*viewHeightMm);
        TouchPoint current1 = notchPoints.get(0);
        current1 = new TouchPoint(current1.x*viewWidthMm, current1.y*viewHeightMm);
        TouchPoint vCurrent = new TouchPoint(
                current1.x - currentCenter.x,
                current1.y - currentCenter.y);
        double angle = TouchTokenRecognizer.angleBetweenVectors(vCurrent, vTransformedTemplate);
        return angle;
    }

    public TouchPoint getTokenCenter(TokenTemplate transformedTemplate, ArrayList<TouchPoint> notchPoints) {
        TouchPoint centerTemplate = new TouchPoint(transformedTemplate.getCenter().x, transformedTemplate.getCenter().y);
        TouchPoint pt1Template = new TouchPoint(transformedTemplate.getPoints().get(0).x, transformedTemplate.getPoints().get(0).y);
        TouchPoint pt2Template = new TouchPoint(transformedTemplate.getPoints().get(1).x, transformedTemplate.getPoints().get(1).y);
        TouchPoint v1CenterTemplate = new TouchPoint(centerTemplate.x - pt1Template.x, centerTemplate.y - pt1Template.y);
        TouchPoint v12Template = new TouchPoint(pt2Template.x - pt1Template.x, pt2Template.y - pt1Template.y);

        double angle21Center = TouchTokenRecognizer.angleBetweenVectors(v12Template, v1CenterTemplate);

        TouchPoint pt1Token = notchPoints.get(0);
        pt1Token = new TouchPoint(pt1Token.x*viewWidthMm, pt1Token.y*viewHeightMm);
        TouchPoint pt2Token = notchPoints.get(1);
        pt2Token = new TouchPoint(pt2Token.x*viewWidthMm, pt2Token.y*viewHeightMm);
        TouchPoint v12Token = new TouchPoint(pt2Token.x - pt1Token.x, pt2Token.y - pt1Token.y);
        double x12Token = TouchTokenRecognizer.angleBetweenVectors(new TouchPoint(1, 0), v12Token);

        double length1Center =
                Math.sqrt((centerTemplate.x - pt1Template.x) * (centerTemplate.x - pt1Template.x) +
                        (centerTemplate.y - pt1Template.y) * (centerTemplate.y - pt1Template.y));

        TouchPoint centerToken = new TouchPoint(length1Center, 0);
        centerToken.setLocation(centerToken.x + pt1Token.getX(), centerToken.y + pt1Token.getY());
        double angle = x12Token + angle21Center;
        TouchTokenRecognizer.rotateBy(centerToken, pt1Token, angle, centerToken);
        centerToken.setLocation(centerToken.x/viewWidthMm, centerToken.y/viewHeightMm);
        return centerToken;
    }

}
